#
# System Interface Library for games
# Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
# Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
# See the file COPYING.txt for details.
#
# build/common/toolchain.mk: Toolchain configuration for the SIL build system.
#

#
# This file attempts to determine the toolchain in use and set up the
# compiler variables required by the rest of the build system.
#
# On input, the following variables should be set (typically by the
# system-specific build.mk file):
#
#    $(CC):  C/Objective-C/assembly compiler command.  This should be a
#       single word, the name of the program to execute.
#
#    $(CXX):  C++ compiler command.  This is optional; if not set, it will
#       be determined from $(CC).
#
#    $(AR_RC), $(RANLIB):  Commands for working with object archives, as
#       used by rules.mk.  These are optional; if not set, they weill be
#       determined from $(CC) assuming a Unix-like system.
#
#    $(SYS_CFLAGS), $(SYS_CXXFLAGS), $(SYS_OBJCFLAGS), $(SYS_ASFLAGS),
#    $(SYS_LDFLAGS):  System-specific flags for each compiler/linker mode
#       (respectively: C compiler, C++ compiler, Objective-C compiler,
#       assembler, linker). These will be appended to the end of the common
#       flag list on the compile or link command line.  These may be empty
#       if no system-specific flags are required.
#
#    $(OBJECT_EXT), $(LIBRARY_EXT):  These should be defined as required
#       for rules.mk, but they are not used directly by this file.
#
# On output, the following variables will be set:
#
#    $(CPP), $(CXX), $(BASE_CFLAGS), $(BASE_CXXFLAGS), $(BASE_OBJCFLAGS),
#    $(CFLAG_DEFINE), $(CFLAG_INCLUDE_DIR), $(CFLAG_INCLUDE_FILE):  These
#       variables will be defined as required for sources.mk.
#
#    $(CFLAG_STD_C99):  A compiler flag which selects the C99 language
#       standard for compiling C and Objective-C sources.  This flag will
#       include support for C11-style anonymous unions.  All SIL sources
#       are built with this flag; you can also include it in the CFLAGS or
#       OBJCFLAGS settings for your own modules.
#
#    $(CFLAG_STD_CXX11):  A compiler flag which selects the C++11 language
#       standard for compiling C++ sources.  All SIL sources are built with
#       this flag; you can also include it in the CXXFLAGS setting for your
#       own modules.
#
#    $(CC_TYPE):  A single word indicating the compiler type.  This will
#       be one of:
#          clang (Clang compiler)
#          gcc   (GCC or compatible compilers)
#          icc   (Intel C++ compiler)
#       If the compiler type is not recognized, "gcc" will be assumed.
#
#    $(CC_VERSION):  A single word indicating the compiler version, or
#       nothing if the compiler version cannot be determined.
#
#    $(CC_VERSION_MAJOR), $(CC_VERSION_MINOR):  The first and second
#       dot-separated parts of $(CC_VERSION), if any.
#
#    $(CC_ARCH):  The target architecture of the compiler, if known.
#       Currently one of: arm mips x86 x86_64 unknown
#

###########################################################################
###########################################################################

# Extract the compiler directory and base command name.  If the base name
# contains any hyphens, take everything up to and including the last hyphen
# as a common prefix to be prepended to other command names.

_SIL_TOOLCHAIN_DIR = $(if $(filter ./%,$(CC)),$(dir $(CC)),$(if $(filter ./,$(dir $(CC))),,$(dir $(CC))))
_SIL_TOOLCHAIN_CC_BASENAME = $(notdir $(CC))
_SIL_TOOLCHAIN_CC_UNVERSIONED = $(shell echo 'x$(_SIL_TOOLCHAIN_CC_BASENAME)' | sed -e 's/^x//' -e 's/-[0-9.]*$$//')
_SIL_TOOLCHAIN_CC_BASE = $(lastword $(subst -,$(preserve-space) $(preserve-space),$(_SIL_TOOLCHAIN_CC_UNVERSIONED)))
_SIL_TOOLCHAIN_PREFIX = $(_SIL_TOOLCHAIN_DIR)$(patsubst %-$(_SIL_TOOLCHAIN_CC_BASE),%-,$(filter %-$(_SIL_TOOLCHAIN_CC_BASE),$(_SIL_TOOLCHAIN_CC_UNVERSIONED)))
_SIL_TOOLCHAIN_SUFFIX = $(patsubst $(_SIL_TOOLCHAIN_CC_BASE)-%,-%,$(filter $(_SIL_TOOLCHAIN_CC_BASE)-%,$(_SIL_TOOLCHAIN_CC_BASENAME)))


# Determine the compiler type.

ifeq ($(_SIL_TOOLCHAIN_CC_BASE),clang)
    CC_TYPE = clang
else ifeq ($(_SIL_TOOLCHAIN_CC_BASE),icc)
    CC_TYPE = icc
else ifeq ($(_SIL_TOOLCHAIN_CC_BASE),gcc)
    CC_TYPE = gcc
else
    _SIL_TOOLCHAIN_CC_VERSION_TEXT := $(shell '$(CC)' --version 2>&1)
    ifneq (,$(filter clang LLVM,$(_SIL_TOOLCHAIN_CC_VERSION_TEXT)))
        CC_TYPE = clang
    else ifneq (,$(filter gcc,$(_SIL_TOOLCHAIN_CC_VERSION_TEXT)))
        CC_TYPE = gcc
    else
        $(warning *** Can't determine compiler type, assuming GCC.)
        $(warning *** If this is incorrect, edit: $(lastword $(MAKEFILE_LIST)))
        CC_TYPE = gcc
    endif
endif


# Determine the compiler version.

ifeq ($(CC_TYPE),clang)
    CC_VERSION := $(shell '$(CC)' --version 2>&1 | grep '^clang version ' | head -n1 | sed -e 's/^clang version \([^ ]*\).*/\1/')
else ifeq ($(CC_TYPE),gcc)
    CC_VERSION := $(lastword $(shell '$(CC)' --version 2>&1 | head -n1))
else ifeq ($(CC_TYPE),icc)
    CC_VERSION := $(lastword $(shell '$(CC)' -v 2>&1 | head -n1))
endif

CC_VERSION_MAJOR := $(word 1,$(subst ., ,$(CC_VERSION)))
CC_VERSION_MINOR := $(word 2,$(subst ., ,$(CC_VERSION)))


# Set tool pathnames.  We'd normally use ?= here to avoid overriding
# values set by the calling Makefile, but Make provides default values for
# some of these variables, so we have to explicitly check for those and
# override them.

ifeq ($(CC_TYPE),clang)
    _SIL_TOOLCHAIN_CXX = $(_SIL_TOOLCHAIN_PREFIX)clang++$(_SIL_TOOLCHAIN_SUFFIX)
else ifeq ($(CC_TYPE),icc)
    _SIL_TOOLCHAIN_CXX = $(_SIL_TOOLCHAIN_PREFIX)icpc$(_SIL_TOOLCHAIN_SUFFIX)
else ifeq ($(_SIL_TOOLCHAIN_CC_BASE),cc)
    _SIL_TOOLCHAIN_CXX = $(_SIL_TOOLCHAIN_PREFIX)c++$(_SIL_TOOLCHAIN_SUFFIX)
else
    _SIL_TOOLCHAIN_CXX = $(_SIL_TOOLCHAIN_PREFIX)$(patsubst %cc,%,$(_SIL_TOOLCHAIN_CC_BASE))++$(_SIL_TOOLCHAIN_SUFFIX)
endif

ifneq (,$(filter undefined default,$(origin CXX)))
    CXX = $(_SIL_TOOLCHAIN_CXX)
endif

ifneq (,$(filter undefined default,$(origin AR_RC)))
    AR_RC = $(_SIL_TOOLCHAIN_PREFIX)ar rc
endif

ifneq (,$(filter undefined default,$(origin RANLIB)))
    RANLIB = $(_SIL_TOOLCHAIN_PREFIX)ranlib
endif


# Set flag variables used by build rules.

CFLAG_PREPROCESS   = -E
CFLAG_DEFINE       = -D
CFLAG_INCLUDE_DIR  = -I
CFLAG_INCLUDE_FILE = -include $(preserve-trailing-space)
# Default to gnu99 instead of c99 for C11-style anonymous union support.
# Recent versions of GCC and Clang allow anonymous unions with -std=c99,
# so we rewrite this to -std=c99 for those compilers.
CFLAG_STD_C99      = -std=gnu99
CFLAG_STD_CXX11    = -std=c++11


# Determine the compiler's target architecture.

CC_ARCH := $(strip \
    $(if $(shell echo __arm__ | $(CC) $(SYS_CFLAGS) $(CFLAGS) $(CFLAG_PREPROCESS) - | grep '^1'),arm,\
    $(if $(shell echo __mips__ | $(CC) $(SYS_CFLAGS) $(CFLAGS) $(CFLAG_PREPROCESS) - | grep '^1'),mips,\
    $(if $(shell echo __i386__ | $(CC) $(SYS_CFLAGS) $(CFLAGS) $(CFLAG_PREPROCESS) - | grep '^1'),x86,\
    $(if $(shell echo __x86_64__ | $(CC) $(SYS_CFLAGS) $(CFLAGS) $(CFLAG_PREPROCESS) - | grep '^1'),x86_64,\
    unknown)))))


# Set base compiler/linker flags for all compiler types.

BASE_CFLAGS = \
    -O3 -pipe -g \
    -fno-stack-protector -fno-strict-overflow \
    -Wall -Wmissing-declarations -Wpointer-arith -Wshadow -Wstrict-prototypes \
    -Wwrite-strings \
    $(if $(filter 1,$(WARNINGS_AS_ERRORS)),-Werror)

BASE_CXXFLAGS = \
    -O3 -pipe -g \
    $(if $(filter 1,$(SIL_ENABLE_CXX_EXCEPTIONS)),-fexceptions,-fno-exceptions) \
    -fsigned-char -fno-stack-protector -fno-strict-overflow \
    -Wall -Wpointer-arith -Wshadow \
    $(if $(filter 1,$(WARNINGS_AS_ERRORS)),-Werror)

BASE_OBJCFLAGS = \
    -O3 -pipe -g \
    -fno-stack-protector -fno-strict-overflow \
    -Wall -Wmissing-declarations -Wpointer-arith -Wshadow -Wstrict-prototypes \
    -Wwrite-strings \
    $(if $(filter 1,$(WARNINGS_AS_ERRORS)),-Werror)

BASE_ASFLAGS = # no flags

BASE_LDFLAGS = \
    $(if $(filter 1,$(SIL_ENABLE_CXX_EXCEPTIONS)),-fexceptions,-fno-exceptions)


# Append compiler-specific flags.

ifeq ($(CC_TYPE),gcc)
    BASE_LDFLAGS += $(if $(filter 1,$(SIL_ENABLE_CXX_EXCEPTIONS)),,-static-libgcc -static-libstdc++)
    ifneq ($(or $(filter-out 1 2 3 4,$(CC_VERSION_MAJOR)),\
                $(and $(filter 4,$(CC_VERSION_MAJOR)),\
                      $(filter-out 0 1 2 3 4 5 6,$(CC_VERSION_MINOR)))),)
        CFLAG_STD_C99   = -std=c99
        BASE_CFLAGS    += -Wunused-local-typedefs
        BASE_CXXFLAGS  += -Wunused-local-typedefs
        BASE_OBJCFLAGS += -Wunused-local-typedefs

        # GCC 4.7 and later versions report numerous "may be uninitialized"
        # warnings.  These are widely[1][2][3] considered useless and ignored
        # due to a high false-positive rate.
        #    [1] http://lkml.org/lkml/2013/3/25/347
        #    [2] https://bugs.webkit.org/show_bug.cgi?id=119835
        #    [3] http://gcc.gnu.org/bugzilla/show_bug.cgi?id=55644
        #
        # In the specific case of SIL, GCC 4.8 reports "may be uninitialized"
        # for libvorbis code that looks roughly like:
        #     int maxval[8], i;
        #     for (i = 0; i < count; i++) {
        #         maxval[i] = get_maxval(i);
        #     }
        #     for (i = 0; i < count; i++) {
        #         do_something_with(maxval[i]);  // "may be uninitialized"
        #     }
        #
        # As with other projects, we treat these warnings as useless and
        # disable them.  The Clang compiler seems to do a better job of
        # accurately reporting uninitialized-data errors.
        BASE_CFLAGS    += -Wno-maybe-uninitialized
        BASE_CXXFLAGS  += -Wno-maybe-uninitialized
        BASE_OBJCFLAGS += -Wno-maybe-uninitialized
    endif

else ifeq ($(CC_TYPE),clang)
    ifneq ($(or $(filter-out 1 2 3,$(CC_VERSION_MAJOR)),\
                $(and $(filter 3,$(CC_VERSION_MAJOR)),\
                      $(filter-out 0 1,$(CC_VERSION_MINOR)))),)
        CFLAG_STD_C = -std=c99
    endif

endif


# Append architecture-specific flags.

ifeq ($(CC_ARCH),x86)
    BASE_CFLAGS    += -msse $(if $(filter clang,$(CC_TYPE)),,-mfpmath=sse)
    BASE_CXXFLAGS  += -msse $(if $(filter clang,$(CC_TYPE)),,-mfpmath=sse)
    BASE_OBJCFLAGS += -msse $(if $(filter clang,$(CC_TYPE)),,-mfpmath=sse)
endif


# Append system-specific flags.

BASE_CFLAGS    += $(SYS_CFLAGS)
BASE_CXXFLAGS  += $(SYS_CXXFLAGS)
BASE_OBJCFLAGS += $(SYS_OBJCFLAGS)
BASE_ASFLAGS   += $(SYS_ASFLAGS)
BASE_LDFLAGS   += $(SYS_LDFLAGS)


# Define flags and tools used for coverage analysis.

ifeq ($(CC_TYPE),gcc)
    COVERAGE_CFLAGS = -fprofile-arcs -ftest-coverage -O0
    COVERAGE_LIBS = -lgcov
    ifneq (,$(filter undefined default,$(origin GCOV)))
        GCOV = $(_SIL_TOOLCHAIN_PREFIX)gcov >/dev/null
        GCOV_OPTS = -b -c -l -p
        GCOV_FILE_OPTS = -o "../$(OBJDIR)/`dirname \"$1\"`" "$1"
        GCOV_STDOUT = >/dev/null
    endif

else ifeq ($(CC_TYPE),clang)
    COVERAGE_CFLAGS = -fprofile-arcs -ftest-coverage -O0
    # The library name for Clang varies with the architecture.
    COVERAGE_LIBS = --coverage
    ifneq (,$(filter undefined default,$(origin GCOV)))
        GCOV = $(_SIL_TOOLCHAIN_PREFIX)llvm-cov
        GCOV_OPTS = -b -c
        GCOV_FILE_OPTS = \
            -gcno="../$(OBJDIR)/`echo \"$1\" | sed -e 's|\.[^./]*$$|.gcno|'`" \
            -gcda="../$(OBJDIR)/`echo \"$1\" | sed -e 's|\.[^./]*$$|.gcda|'`" \
            -o "`echo "$1" | sed -e 's|[/.]|-|g'`.out"
    endif

else
    COVERAGE_CFLAGS = $(warning Warning: Coverage flags unknown for this compiler.)
    COVERAGE_LIBS = $(warning Warning: Coverage flags unknown for this compiler.)
    GCOV = : $(error Coverage tool unknown for this compiler)
endif

COVERAGE_CXXFLAGS = $(COVERAGE_CFLAGS)
COVERAGE_OBJCFLAGS = $(COVERAGE_CFLAGS)
COVERAGE_ASFLAGS =

###########################################################################
###########################################################################
