#
# System Interface Library for games
# Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
# Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
# See the file COPYING.txt for details.
#
# build/ios/build.mk: Build control file for iOS.
#

#
# iOS-specific notes:
#
# - $(PROGRAM_VERSION_CODE) is used as the CFBundleVersion key value in
#   the Info.plist file, and it must match /[0-9]+(\.[0-9]+)*/.
#
# - $(SDK_VERSION) and $(TARGET_OS_VERSION), if set, should refer to an
#   iOS version number in "major.minor" form; for example:
#      SDK_VERSION = 6.0
#
# - To add resources to the package, create a target for each resource file
#   or directory named "$(BUNDLE_DIR_ESCAPED)/<path>" and make the "data"
#   target depend on that path.  For example:
#      data: $(BUNDLE_DIR_ESCAPED)/data
#      $(BUNDLE_DIR_ESCAPED)/data: $(TOPDIR)/data $(MAKEFILE_DEPS)
#      <tab>rm -rf '$@'
#      <tab>cp -a '$<' '$@'
#
# - When collecting coverage data, it is necessary to copy the raw coverage
#   data files from the device before running "make gen-coverage".  The
#   data is located in "Library/Application Support/coverage" in the app
#   installation directory, under the same path structure as the absolute
#   path on the build system; for example, if the program's build directory
#   is "/Users/username/MyApp/build", the coverage data would be found in
#   "Library/Application Support/coverage/Users/username/MyApp/build/obj".
#   Copy all files from this "obj" directory on the device to the "obj"
#   directory on the build system, preserving relative pathnames.
#
#   Note that coverage data will only be output in debug mode.
#
# -------------------------------------------------------------------------
#
# To use this file, include it in your Makefile after defining values for
# configuration variables and source lists, but before defining any rules.
# For example:
#
# PROGRAM_NAME = MyProgram
# EXECUTABLE_NAME = my-program
#
# MY_SOURCES = my-program.c
# MY_OBJECTS_C = $(MY_SOURCES:$(TOPDIR)/%.c=$(OBJDIR)/%$(OBJECT_EXT))
# OBJECT_GROUPS += MY
# AUTOGEN_HEADERS += my-date.h
#
# include SIL/build/ios/build.mk
#
# .PHONY: my-date.h  # Force regeneration on every build.
# my-date.h:
# 	date +'#define DATE "%Y-%m-%d %H:%M:%S"' >$@
#
#

###########################################################################
################### iOS-specific configuration options ####################
###########################################################################

#---------------------- Build environment pathnames ----------------------#

# $(DEV_ROOT):  Root path of the iOS development tools.

DEV_ROOT ?= $(XCODE_ROOT)/Platforms/iPhoneOS.platform/Developer


# $(TOOLCHAIN_ROOT):  Root path of the compiler toolchain.
#
# Implementation note: the use of $(shell ls) instead of $(wildcard) is
# deliberate, since $(wildcard) breaks if there are spaces in the pathname.

TOOLCHAIN_ROOT ?= $(or $(shell ls -d '$(XCODE_ROOT)/Toolchains/XcodeDefault.xctoolchain' 2>/dev/null),$(DEV_ROOT))


# $(XCODE_APP):  Installation path of the Xcode application, normally
# "/Applications/Xcode.app".  This is only used to set the default value
# for $(XCODE_ROOT).

XCODE_APP ?= /Applications/Xcode.app


# $(XCODE_ROOT):  Installation path for Xcode developer files, normally
# "Contents/Developer" inside the Xcode.app package.  This is only used to
# set the default values for $(DEV_ROOT) and $(TOOLCHAIN_ROOT).

XCODE_ROOT ?= $(XCODE_APP)/Contents/Developer

#----------------------- Program behavior options ------------------------#

# $(RUN_TESTS):  Set this to 1 to run the test routines instead of the
# main program.

RUN_TESTS ?= 0


# $(USE_FILE_SHARING):  Set this to 1 to enable exporting of save files
# (and their associated screenshots) to the Documents directory, from which
# they can be copied off the device via iTunes file sharing.  Save files
# copied onto the device via iTunes will likewise be imported, overwriting
# any existing save in the game's internal storage.
#
# If GameKit is enabled (see below), the set of exported save files always
# reflects the currently signed-in Game Center player (or the most recently
# signed-in player, if no player is currently signed in).  When a new player
# is detected, the exported save files are replaced with those of the new
# player.  (As a corollary, if you want to copy save files to a new player,
# you have to start the game once with the new player signed in before you
# can copy the saves to the device -- otherwise your copied saves will be
# overwritten when the game detects the new player.)
#
# IMPORTANT:  This functionality should be considered _experimental_!
# The details of the file sharing implementation appear to be undefined
# by Apple documentation; SIL attempts to ensure that internal and
# external data copies are kept in sync without excessive load/save
# overhead, but there may still be a potential for data loss in either
# direction.  Use with care!

USE_FILE_SHARING ?= 0


# $(USE_GAMEKIT):  Set this to 1 to enable Game Center functionality via
# the GameKit framework.  If enabled, achievements will be sent to Game
# Center as well as stored locally; see sysdep/ios/userdata.c for details.
#
# This is generally useless unless you have an Apple developer account and
# associated app record with Game Center enabled.

USE_GAMEKIT ?= 0

#--------------------------- Packaging options ---------------------------#

# $(DEVICE_IPHONE), $(DEVICE_IPAD):  Set these to 1 to enable the program
# to run on the respective device (DEVICE_IPHONE includes both the iPhone
# and the iPod touch).

DEVICE_IPHONE ?= 1
DEVICE_IPAD ?= 1


# $(INFO_PLIST_SOURCE):  Info.plist file into which build-specific values
# will be inserted.  If not set, a standard Info.plist will be used.

INFO_PLIST_SOURCE ?= $(_THISDIR)/Info.plist


# $(PROFILE_NAME), $(PROFILE_FILE):  Use these to select the provisioning
# profile to build with.  $(PROFILE_NAME) specifies a name substring
# matched against the names of all existing profiles with valid
# certificates; $(PROFILE_FILE) can be used to specify a particular
# profile file (*.mobileprovision), and takes precedence if defined.
# If both are left empty, $(PROFILE_NAME) will default to "Team
# Provisioning Profile" for debug builds (when $(DEBUG) is 1), and to
# "Distribution" for non-debug builds.

PROFILE_NAME ?=
PROFILE_FILE ?=


# $(VALIDATE):  If set to 1, the Xcode package validator will be run on
# the generated bundle after it is signed.

VALIDATE ?= 0

###########################################################################
##################### Common pathnames and filenames ######################
###########################################################################

# SDK root path and canonical version.  The pathname could potentially
# contain spaces, so use $(shell ls) instead of $(wildcard) to check for
# existence.
SDK_PREFIX = $(DEV_ROOT)/SDKs/iPhoneOS
ifeq ($(SDK_VERSION),)
override SDK_VERSION := $(shell ls -d '$(SDK_PREFIX)'* 2>/dev/null \
                                | sed -e 's,.*/,,; s/iPhoneOS//; s/\.sdk//' \
                                | sort -n \
                                | tail -n1)
endif
SDK_ROOT := $(SDK_PREFIX)$(SDK_VERSION).sdk

# If we're going to compile anything, make sure the various paths exist.
# Again, we use $(shell ls) instead of $(wildcard) to avoid problems if
# there are spaces in the pathname.

ifneq ($(filter-out help clean spotless,$(or $(MAKECMDGOALS),default)),)

ifeq ($(shell ls -d '$(DEV_ROOT)' 2>/dev/null),)
$(warning --- Failed to find Xcode iOS tools at $$(DEV_ROOT):)
$(warning ---     $(DEV_ROOT))
$(warning --- Rerun make with "XCODE_APP=/path/to/Xcode.app")
$(warning --- or "DEV_ROOT=/path/to/iPhoneOS.platform/Developer".)
$(error Xcode not found)
endif

ifeq ($(shell ls -d '$(TOOLCHAIN_ROOT)/usr/bin' 2>/dev/null),)
$(warning --- Failed to find Xcode toolchain at $$(TOOLCHAIN_ROOT):)
$(warning ---     $(TOOLCHAIN_ROOT))
$(warning --- Rerun make with "TOOLCHAIN_ROOT=/path/to/toolchain".)
$(error Xcode toolchain not found)
endif

ifeq ($(SDK_VERSION),)
$(warning --- Failed to find any iOS SDK under $$(DEV_ROOT):)
$(warning ---     $(DEV_ROOT))
$(warning --- Rerun make with "SDK_ROOT=/path/to/iPhoneOS.sdk".)
$(error iOS SDK not found)
endif

ifeq ($(shell ls -d '$(SDK_ROOT)' 2>/dev/null),)
$(warning --- Failed to find iOS SDK at $$(SDK_ROOT):)
$(warning ---     $(SDK_ROOT))
$(warning --- Rerun make with "SDK_ROOT=/path/to/iPhoneOS.sdk".)
$(error iOS SDK not found)
endif

ifneq ($(filter 1 2 3 4 5,$(firstword $(subst ., ,$(SDK_VERSION)))),)
$(warning --- iOS SDK (version $(SDK_VERSION)) is too old!)
$(warning --- Version 6.0 or later of the SDK is required.)
$(error iOS SDK too old)
endif

endif

# Target OS version (if not already defined).
TARGET_OS_VERSION ?= 5.0

# Profile selection cleanup.
ifneq ($(PROFILE_FILE),)
override PROFILE_NAME :=
else
ifeq ($(PROFILE_NAME),)
override PROFILE_NAME = $(if $(filter 1,$(DEBUG)),Team Provisioning Profile,Distribution)
endif
endif

#-------------------------------------------------------------------------#

# Output bundle (.app) directory name.
BUNDLE_DIR = $(PROGRAM_NAME).app

# The same thing, with spaces escaped for use in rules.
BUNDLE_DIR_ESCAPED = $(shell echo '$(BUNDLE_DIR)' | sed -e 's/ /\\ /g')

# Output package (.ipa) filename.
PACKAGE_FILE = $(PROGRAM_NAME).ipa

###########################################################################
######################### Build environment setup #########################
###########################################################################

_THISDIR := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

include $(_THISDIR)/../common/base.mk
include $(_THISDIR)/../common/config.mk


CC = $(TOOLCHAIN_ROOT)/usr/bin/clang

# Compiler ID to embed in the Info.plist file under the DTCompiler key.
# Doesn't seem to actually be necessary (at least Xcode 4.3.1 doesn't set it).
#CC_ID = com.apple.compilers.llvm.clang.1_0
CC_ID =

# Default flags used by Xcode (excluding those we already set in toolchain.mk).
PLATFORM_FLAGS = -arch armv7 \
                 -gdwarf-2 \
                 -isysroot $(SDK_ROOT) \
                 -mthumb
PLATFORM_CFLAGS = $(PLATFORM_FLAGS) \
                  $(if $(filter 1,$(DEBUG)),,-DNS_BLOCK_ASSERTIONS=1) \
                  -ffast-math \
                  -fmessage-length=0 \
                  -fvisibility=hidden \
                  -miphoneos-version-min=$(TARGET_OS_VERSION)
PLATFORM_CXXFLAGS = $(PLATFORM_CFLAGS) \
                    -fvisibility-inlines-hidden

AR_RC  = $(TOOLCHAIN_ROOT)/usr/bin/ar rc
RANLIB = $(TOOLCHAIN_ROOT)/usr/bin/ranlib

OBJECT_EXT  = .o
LIBRARY_EXT = .a

SYS_CFLAGS    = $(PLATFORM_CFLAGS)
SYS_CXXFLAGS  = $(PLATFORM_CXXFLAGS)
SYS_OBJCFLAGS = $(PLATFORM_CFLAGS)
SYS_ASFLAGS   = $(PLATFORM_FLAGS)
SYS_LDFLAGS   =

# Pre-5.1 Xcode doesn't ship llvm-cov and separates gcov from the toolchain.
ifeq (,$(wildcard $(TOOLCHAIN_ROOT)/usr/bin/llvm-cov))
    GCOV = '$(XCODE_ROOT)/usr/bin/gcov'
    GCOV_OPTS = -b -c -l -p
    GCOV_FILE_OPTS = -o "../$(OBJDIR)/`dirname \"$1\"`" "$1"
    GCOV_STDOUT = >/dev/null
else
    # Xcode's llvm-cov is a bastardization of vanilla llvm-cov and gcov.
    # We can mostly use gcov-style options with it, but it doesn't accept
    # the -l option, so we need to use a wrapper to keep output files for
    # headers separate from each other.
    GCOV = '$(SIL_DIR)/build/macosx/llvm-cov-wrapper.sh' '$(TOOLCHAIN_ROOT)/usr/bin/llvm-cov'
    GCOV_OPTS = -b -c -p
    GCOV_FILE_OPTS = -o "../$(OBJDIR)/`dirname \"$1\"`" "$1"
    GCOV_STDOUT = >/dev/null
endif

include $(_THISDIR)/../common/toolchain.mk


# Program paths and other definitions used only in this file.

FRAMEWORKS = AudioToolbox AVFoundation CoreGraphics CoreMedia CoreText \
             CoreVideo Foundation GameKit OpenGLES QuartzCore UIKit \
             $(EXTRA_FRAMEWORKS)

###########################################################################
################# Source file lists and associated flags ##################
###########################################################################

SIL_SYS_CONFIG_HEADER = ios.h

SIL_SYS_SOURCES = $(SIL_OPENGL_SOURCES) \
                  sysdep/darwin/debug.c \
                  sysdep/darwin/meminfo.c \
                  sysdep/darwin/mutex.c \
                  sysdep/darwin/semaphore.c \
                  sysdep/darwin/time.c \
                  sysdep/ios/dialog.m \
                  sysdep/ios/gamekit.m \
                  sysdep/ios/graphics.c \
                  sysdep/ios/input.c \
                  sysdep/ios/log.m \
                  sysdep/ios/main.m \
                  sysdep/ios/misc.c \
                  sysdep/ios/osk.m \
                  sysdep/ios/sound.m \
                  sysdep/ios/sysfont.c \
                  sysdep/ios/userdata.c \
                  sysdep/ios/util.m \
                  sysdep/ios/view.m \
                  sysdep/macosx/filedata.m \
                  sysdep/macosx/movie.m \
                  sysdep/posix/files.c \
                  sysdep/posix/fileutil.c \
                  sysdep/posix/ioqueue.c \
                  sysdep/posix/misc.c \
                  sysdep/posix/thread.c \
                  test/sysdep/darwin/time.c \
                  test/sysdep/ios/graphics.c \
                  test/sysdep/ios/util.c \
                  test/sysdep/posix/files.c \
                  test/sysdep/posix/fileutil.c \
                  test/sysdep/posix/internal.c \
                  test/sysdep/posix/ioqueue.c \
                  test/sysdep/posix/misc.c \
                  test/sysdep/posix/thread.c \
                  test/sysdep/posix/userdata.c

SIL_SYS_FLAGS = $(call _if-define,RUN_TESTS) \
                $(call _if-define,USE_FILE_SHARING,SIL_PLATFORM_IOS_USE_FILE_SHARING) \
                $(call _if-define,USE_GAMEKIT,SIL_PLATFORM_IOS_USE_GAMEKIT) \
                $(CFLAG_DEFINE)_POSIX_C_SOURCE=201112L \
                $(CFLAG_DEFINE)_DARWIN_C_SOURCE

# Force ARM rather than Thumb instruction mode for SIL and client sources
# because it seems to provide better performance.
SIL_EXTRA_FLAGS += -marm

include $(_THISDIR)/../common/sources.mk

# We meed to be able to recognize inf/nan in strformat() and userdata code,
# but GCC optimizes out even explicit calls to isinf() and isnan() when
# -ffast-math is in use.
ifneq ($(filter gcc-4.2 llvm-gcc-4.2,$(COMPILER)),)
$(subst $(TOPDIR)/,$(OBJDIR)/,$(SIL_DIR)/src/utility/strformat$(OBJECT_EXT)): SIL_CFLAGS += -fno-fast-math
$(subst $(TOPDIR)/,$(OBJDIR)/,$(SIL_DIR)/src/sysdep/ios/userdata$(OBJECT_EXT)): SIL_CFLAGS += -fno-fast-math
$(subst $(TOPDIR)/,$(OBJDIR)/,$(SIL_DIR)/src/sysdep/test/userdata$(OBJECT_EXT)): SIL_CFLAGS += -fno-fast-math
endif

# Special case for GameKit deprecation warnings which we ignore.
$(subst $(TOPDIR)/,$(OBJDIR)/,$(SIL_DIR)/src/sysdep/ios/gamekit$(OBJECT_EXT)): SIL_OBJCFLAGS += -Wno-deprecated-declarations

###########################################################################
############################### Build rules ###############################
###########################################################################

SIL_DEFAULT_TARGET = all

SIL_HELP_all          = 'build the application package ($(PACKAGE_FILE))'
SIL_HELP_binary       = 'build just the binary'
SIL_HELP_gen-coverage = 'merge coverage data (pre-copied from device) to coverage.out'
SIL_HELP_clean        = 'remove all intermediate files and dependency data'
SIL_HELP_spotless     = 'remove all generated files, including .ipa and bundle dir'
define SIL_HELPTRAILER
echo ''
echo 'Using Xcode installation path (XCODE_APP):   $(XCODE_APP)'
echo 'Building with SDK version (SDK_VERSION):     $(or $(SDK_VERSION),<not found>)'
echo 'Building for OS version (TARGET_OS_VERSION): $(or $(TARGET_OS_VERSION),<unknown>)'
echo ''
endef

include $(_THISDIR)/../common/rules.mk

#-------------------------------------------------------------------------#

.PHONY: all binary clean spotless

all: package

binary: $(BUNDLE_DIR_ESCAPED)/$(EXECUTABLE_NAME)

clean:
	$(ECHO) 'Removing object files'
	$(Q)rm -rf obj
	$(ECHO) 'Removing generated files'
	$(Q)rm -f $(EXECUTABLE_NAME).xcent
	$(ECHO) 'Removing temporary files'
	$(Q)rm -rf .covtmp
	$(SIL_RECIPE_clean)

spotless: clean
	$(ECHO) 'Removing application files'
	$(Q)rm -rf '$(BUNDLE_DIR)' '$(BUNDLE_DIR).dSYM' '$(PACKAGE_FILE)'
	$(ECHO) 'Removing coverage data'
	$(Q)rm -f coverage.out
	$(SIL_RECIPE_spotless)

#-------------------------------------------------------------------------#

.PHONY: package
package: bundle
	$(ECHO) 'Packaging $(PACKAGE_FILE)'
	@# PackageApplication seems to need an absolute path for -o.
	$(Q)'$(DEV_ROOT)/usr/bin/PackageApplication' '$(BUNDLE_DIR)' -o '$(abspath .)/$(PACKAGE_FILE)'

.PHONY: bundle
ifeq ($(VALIDATE),1)
bundle: validated-bundle
else
bundle: signed-bundle
endif

.PHONY: validated-bundle
validated-bundle: signed-bundle
	$(ECHO) 'Validating $(BUNDLE_DIR)'
	$(Q)'$(DEV_ROOT)/usr/bin/Validation' '$(BUNDLE_DIR)'

.PHONY: signed-bundle
signed-bundle: binary data
	$(ECHO) 'Signing $(BUNDLE_DIR)'
	$(Q)rm -rf '$(BUNDLE_DIR)'/_CodeSignature '$(BUNDLE_DIR)'/CodeResources
	$(Q)IDENTITY=`'$(_THISDIR)/signing-helper.pl' \
	    $(if $(PROFILE_NAME),--profile='$(subst ','\'',$(PROFILE_NAME))') \
	    $(if $(PROFILE_FILE),--profile-file='$(subst ','\'',$(PROFILE_FILE))') \
	    $(if $(filter 1,$(V)),--verbose) \
	    '$(PACKAGE_NAME)' \
	    '$(BUNDLE_DIR)/embedded.mobileprovision' \
	    $(EXECUTABLE_NAME).xcent`; \
	test -n "$$IDENTITY" && \
	CODESIGN_ALLOCATE='$(DEV_ROOT)/usr/bin/codesign_allocate' /usr/bin/codesign \
	    --force \
	    --sign "$$IDENTITY" \
	    --resource-rules='$(BUNDLE_DIR)/ResourceRules.plist' \
	    --entitlements $(EXECUTABLE_NAME).xcent \
	    '$(BUNDLE_DIR)'

#-------------------------------------------------------------------------#

$(BUNDLE_DIR_ESCAPED)/$(EXECUTABLE_NAME): $(sort $(ALL_OBJECTS)) always-relink
	$(ECHO) 'Linking $@'
	@# As with the OSX build rules, we can't use $(@D) because it breaks
	@# if $(BUNDLE_DIR) contains spaces.
	$(Q)mkdir -p '$(BUNDLE_DIR)'
	$(Q)$(CXX) \
	    -arch armv7 \
	    -isysroot $(SDK_ROOT) \
	    $(BASE_LDFLAGS) $(ALL_LDFLAGS) $(LDFLAGS) \
	    $(ALL_OBJECTS) \
	    -dead_strip \
	    -miphoneos-version-min='$(TARGET_OS_VERSION)' \
	    $(foreach I,$(FRAMEWORKS),-framework $I) \
            $(BASE_LIBS) $(ALL_LIBS) $(LIBS) \
	    -o '$@'
	$(ECHO) 'Generating symbol data in $(BUNDLE_DIR).dSYM'
	$(Q)$(TOOLCHAIN_ROOT)/usr/bin/dsymutil '$@' -o '$(BUNDLE_DIR).dSYM'

# If we leave an existing binary in place because no objects have been
# updated, the codesign tool will theoretically rewrite the signature in
# the existing binary; however, experience has shown that iOS devices
# don't always accept such re-signed binaries, so we use this dummy rule
# to unconditionally force a relink.
.PHONY: always-relink
always-relink:

#-------------------------------------------------------------------------#

.PHONY: data
data: $(BUNDLE_DIR_ESCAPED)/Info.plist \
      $(BUNDLE_DIR_ESCAPED)/PkgInfo \
      $(BUNDLE_DIR_ESCAPED)/ResourceRules.plist \
      $(BUNDLE_DIR_ESCAPED)/en.lproj/SIL.strings \
      $(if $(filter -DINCLUDE_TESTS,$(SIL_CFLAGS)),$(BUNDLE_DIR_ESCAPED)/testdata)

# Always regenerate Info.plist so we pick up the proper $(PROGRAM_VERSION_CODE)
# in case that's an autogenerated value.
.PHONY: $(BUNDLE_DIR_ESCAPED)/Info.plist
$(BUNDLE_DIR_ESCAPED)/Info.plist:
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)'
	$(Q)perl \
	    -e '$$xcode = `"$(XCODE_ROOT)/usr/bin/xcodebuild" -version`;' \
	    -e '$$xcode =~ /^Xcode (\d+)\.(\d+)(?:\.(\d+))?/m or die "Bad output from xcodebuild -version:\n$$xcode";' \
	    -e '$$xcode_version = sprintf("%02d%d%d", $$1, $$2, $$3);' \
	    -e '($$xcode_build) = ($$xcode =~ /^Build.*\s(\S+)/m);' \
	    -e '$$sdk = `"$(XCODE_ROOT)/usr/bin/xcodebuild" -version -sdk iphoneos$(SDK_VERSION)`;' \
	    -e '$$sdk =~ /^ProductBuildVersion:\s+(\S+)/m or die "Bad output from xcodebuild -version -sdk iphoneos$(SDK_VERSION):\n$$sdk";' \
	    -e '$$sdk_build = $$1;' \
	    -e '($$add{'BuildMachineOSBuild'}) = (`sw_vers -buildVersion` =~ /(.+)/);' \
	    -e '$$add{'CFBundleResourceSpecification'} = "ResourceRules.plist";' \
	    -e '$$add{'DTCompiler'} = "$(CC_ID)";' \
	    -e '$$add{'DTPlatformBuild'} = "$$sdk_build";' \
	    -e '$$add{'DTPlatformName'} = "iphoneos";' \
	    -e '$$add{'DTPlatformVersion'} = "$(SDK_VERSION)";' \
	    -e '$$add{'DTSDKBuild'} = "$$sdk_build";' \
	    -e '$$add{'DTSDKName'} = "iphoneos$(SDK_VERSION)";' \
	    -e '$$add{'DTXcode'} = "$$xcode_version";' \
	    -e '$$add{'DTXcodeBuild'} = "$$xcode_build";' \
	    -e '$$add{'MinimumOSVersion'} = "$(TARGET_OS_VERSION)";' \
	    -e '$$add_str = join("", map {sprintf("\t<key>%s</key>\n\t<string>%s</string>\n", $$_, $$add{$$_})} sort(keys(%add)));' \
	    -e '$$add_str .= "\t<key>UIFileSharingEnabled</key>\n\t<true/>\n" if "$(USE_FILE_SHARING)";' \
	    -e '$$add_str .= "\t<key>UIViewControllerBasedStatusBarAppearance</key>\n\t<false/>\n" if "$(firstword $(subst ., ,$(SDK_VERSION)))" >= 7;' \
	    -e '$$add_str .= "\t<key>CFBundleSupportedPlatforms</key>\n\t<array><string>iPhoneOS</string></array>\n";' \
	    -e '@devices = ();' \
	    -e 'push @devices, 1 if "$(DEVICE_IPHONE)";' \
	    -e 'push @devices, 2 if "$(DEVICE_IPAD)";' \
	    -e '$$add_str .= "\t<key>UIDeviceFamily</key>\n\t<array>\n"' \
	    -e '           . join("", map {"\t\t<integer>$$_</integer>\n"} @devices)' \
	    -e '           . "\t</array>\n";' \
	    -e 'undef $$/; $$_ = <STDIN>;' \
	    -e 's/\$$\{EXECUTABLE_NAME\}/$(EXECUTABLE_NAME)/g;' \
	    -e 's/\$$\{IPHONEOS_DEPLOYMENT_TARGET\}/$(TARGET_OS_VERSION)/g;' \
	    -e 's/\$$\{PRODUCT_NAME\}/$(PROGRAM_NAME)/g;' \
	    -e 's/(CFBundleIdentifier.*\n.*)<string>[^<]*/$$1<string>$(PACKAGE_NAME)/g;' \
	    -e 's/(CFBundleShortVersionString.*\n.*)<string>[^<]*/$$1<string>$(PROGRAM_VERSION)/g if "$(PROGRAM_VERSION)" ne "";' \
	    -e 's/(CFBundleVersion.*\n.*)<string>[^<]*/$$1<string>$(PROGRAM_VERSION_CODE)/g;' \
	    -e 's|\n(</dict>)|\n$${add_str}$$1|;' \
	    -e 'print;' \
	    <'$(INFO_PLIST_SOURCE)' >'$@' \
	    && plutil -convert binary1 '$@' \
	    || rm -f '$@'

$(BUNDLE_DIR_ESCAPED)/PkgInfo: $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)'
	$(Q)printf 'APPL????' >'$@'

$(BUNDLE_DIR_ESCAPED)/ResourceRules.plist: $(_THISDIR)/ResourceRules.plist $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)'
	$(Q)cp -p '$<' '$@'

$(BUNDLE_DIR_ESCAPED)/en.lproj/SIL.strings: $(_THISDIR)/en.lproj/SIL.strings $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)/en.lproj'
	$(Q)$(SIL_DIR)/build/macosx/copystrings.sh \
	    --validate \
	    --inputencoding utf-8 \
	    --outputencoding binary \
	    --outdir '$(BUNDLE_DIR)/en.lproj' \
	    '$<'

$(BUNDLE_DIR_ESCAPED)/testdata: $(SIL_DIR)/testdata $(MAKEFILE_DEPS)
	$(ECHO) 'Copying $@ from $<'
	@# Delete the old directory so removed files don't hang around.
	$(Q)rm -rf '$@'
	$(Q)mkdir -p '$@'
	$(Q)cp -a '$<'/* '$@/'

###########################################################################
###########################################################################
