/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/mutex.c: Mutual-exclusion lock (mutex) handling.
 */

#include "src/base.h"
#include "src/mutex.h"
#include "src/sysdep.h"
#include "src/utility/id-array.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Array of created mutex objects. */
static IDArray mutexes = ID_ARRAY_INITIALIZER(10);

/**
 * VALIDATE_MUTEX:  Validate the mutex ID passed to a mutex
 * manipulation routine, and store the corresponding pointer in the
 * variable "mutex".  If the mutex pointer is invalid, the
 * "error_return" statement is executed; this may consist of multiple
 * statements, but must include a "return" to exit the function.
 */
#define VALIDATE_MUTEX(id,mutex,error_return) \
    ID_ARRAY_VALIDATE(&mutexes, (id), SysMutexID, mutex, \
                      DLOG("Mutex ID %d is invalid", __id); error_return)

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int mutex_create(MutexState initial_state)
{
    if (UNLIKELY(initial_state != MUTEX_LOCKED
                 && initial_state != MUTEX_UNLOCKED)) {
        DLOG("Invalid initial_state: %d", (int)initial_state);
        return 0;
    }

    SysMutexID mutex = sys_mutex_create((int)initial_state);
    if (!mutex) {
        return 0;
    }

    int id = id_array_register(&mutexes, (void *)mutex);
    if (UNLIKELY(!id)) {
        DLOG("Failed to store new mutex in array");
        sys_mutex_destroy(mutex);
        return 0;
    }

    return id;
}

/*-----------------------------------------------------------------------*/

void mutex_destroy(int mutex)
{
    if (mutex) {
        SysMutexID sys_mutex;
        VALIDATE_MUTEX(mutex, sys_mutex, return);
        sys_mutex_destroy(sys_mutex);
        id_array_release(&mutexes, mutex);
    }
}

/*-----------------------------------------------------------------------*/

void mutex_lock(int mutex)
{
    SysMutexID sys_mutex;
    VALIDATE_MUTEX(mutex, sys_mutex, return);
    sys_mutex_lock(sys_mutex);
}

/*-----------------------------------------------------------------------*/

int mutex_lock_timeout(int mutex, float timeout)
{
    SysMutexID sys_mutex;
    VALIDATE_MUTEX(mutex, sys_mutex, return 0);

    /* We check !(timeout >= 0) instead of (timeout < 0) so that NaNs will
     * fail the test. */
    if (UNLIKELY(!(timeout >= 0))) {
        DLOG("Invalid timeout: %g", timeout);
        return 0;
    }

    return sys_mutex_lock_timeout(sys_mutex, timeout);
}

/*-----------------------------------------------------------------------*/

void mutex_unlock(int mutex)
{
    SysMutexID sys_mutex;
    VALIDATE_MUTEX(mutex, sys_mutex, return);
    sys_mutex_unlock(sys_mutex);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
