/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/sdl/input.c: Input device interface for SDL.
 */

/*
 * SDL does not have the concept of a "stick" on a joystick, instead
 * collecting all analog inputs under the single label "axis".  We assume
 * that on any joystick device with a stick, the Y axis of the stick will
 * have the next consecutive axis number after the X axis, and we use the
 * X axis index as the joystick index.  For example, on a joystick with
 * two sticks where one stick has axes X=0 and Y=1 and the other has X=2
 * and Y=3 (such as the PlayStation 3 gamepad), the first stick will have
 * index 0 and the second stick will have index 2.  sys_input_info() will
 * return 4(*) "sticks" in this case, but sticks 1 and 3 will not provide
 * meaningful data.
 * (*) Actually, the PS3 gamepad has 20 axes, with the other 16 being
 *     the pressure sensors on the various buttons, so sys_input_info()
 *     will return 20 "sticks".
 *
 * In the case of (arguably poorly-designed) joysticks that do not assign
 * axes as described above, it is still possible to obtain the individual
 * axis values by calling input_joystick_read_stick() twice, passing NULL
 * for the y_ret parameter.  For example, if a stick had X axis 2 and
 * Y axis 5, the stick position could be obtained as follows:
 *     float x, y;
 *     input_joystick_read_stick(2, &x, NULL);
 *     input_joystick_read_stick(5, &y, NULL);
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/graphics.h"
#include "src/input.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/sdl/internal.h"
#include "src/time.h"
#include "src/utility/utf8.h"

#include <SDL.h>
#if SDL_VERSION_ATLEAST(2,0,0)
# include <SDL_gamecontroller.h>
# include <SDL_haptic.h>
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Macro mapping SDLK_* symbols from SDL 2.x so they're usable as array
 * indices. */
#if SDL_VERSION_ATLEAST(2,0,0)
# define SIL_SDLK(key)  ((key) & SDLK_SCANCODE_MASK ? 256 + ((key) & ~SDLK_SCANCODE_MASK) : (key))
#else
# define SIL_SDLK(key)  (key)
#endif

/* Mapping from SDL key symbols to KEY_* symbols (defined as uint8_t to
 * save space, since currently all KEY_* values fit in 8 bits). */
static const uint8_t sdlkey_map[] = {
    [SIL_SDLK(SDLK_a)           ] = KEY_A,
    [SIL_SDLK(SDLK_b)           ] = KEY_B,
    [SIL_SDLK(SDLK_c)           ] = KEY_C,
    [SIL_SDLK(SDLK_d)           ] = KEY_D,
    [SIL_SDLK(SDLK_e)           ] = KEY_E,
    [SIL_SDLK(SDLK_f)           ] = KEY_F,
    [SIL_SDLK(SDLK_g)           ] = KEY_G,
    [SIL_SDLK(SDLK_h)           ] = KEY_H,
    [SIL_SDLK(SDLK_i)           ] = KEY_I,
    [SIL_SDLK(SDLK_j)           ] = KEY_J,
    [SIL_SDLK(SDLK_k)           ] = KEY_K,
    [SIL_SDLK(SDLK_l)           ] = KEY_L,
    [SIL_SDLK(SDLK_m)           ] = KEY_M,
    [SIL_SDLK(SDLK_n)           ] = KEY_N,
    [SIL_SDLK(SDLK_o)           ] = KEY_O,
    [SIL_SDLK(SDLK_p)           ] = KEY_P,
    [SIL_SDLK(SDLK_q)           ] = KEY_Q,
    [SIL_SDLK(SDLK_r)           ] = KEY_R,
    [SIL_SDLK(SDLK_s)           ] = KEY_S,
    [SIL_SDLK(SDLK_t)           ] = KEY_T,
    [SIL_SDLK(SDLK_u)           ] = KEY_U,
    [SIL_SDLK(SDLK_v)           ] = KEY_V,
    [SIL_SDLK(SDLK_w)           ] = KEY_W,
    [SIL_SDLK(SDLK_x)           ] = KEY_X,
    [SIL_SDLK(SDLK_y)           ] = KEY_Y,
    [SIL_SDLK(SDLK_z)           ] = KEY_Z,
    [SIL_SDLK(SDLK_0)           ] = KEY_0,
    [SIL_SDLK(SDLK_1)           ] = KEY_1,
    [SIL_SDLK(SDLK_2)           ] = KEY_2,
    [SIL_SDLK(SDLK_3)           ] = KEY_3,
    [SIL_SDLK(SDLK_4)           ] = KEY_4,
    [SIL_SDLK(SDLK_5)           ] = KEY_5,
    [SIL_SDLK(SDLK_6)           ] = KEY_6,
    [SIL_SDLK(SDLK_7)           ] = KEY_7,
    [SIL_SDLK(SDLK_8)           ] = KEY_8,
    [SIL_SDLK(SDLK_9)           ] = KEY_9,
    [SIL_SDLK(SDLK_TAB)         ] = KEY_TAB,
    [SIL_SDLK(SDLK_RETURN)      ] = KEY_ENTER,
    [SIL_SDLK(SDLK_SPACE)       ] = KEY_SPACE,
    [SIL_SDLK(SDLK_EXCLAIM)     ] = KEY_EXCLAMATION,
    [SIL_SDLK(SDLK_QUOTEDBL)    ] = KEY_QUOTE,
    [SIL_SDLK(SDLK_HASH)        ] = KEY_SHARP,
    [SIL_SDLK(SDLK_DOLLAR)      ] = KEY_DOLLAR,
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_PERCENT)     ] = KEY_PERCENT,
#endif
    [SIL_SDLK(SDLK_AMPERSAND)   ] = KEY_AMPERSAND,
    [SIL_SDLK(SDLK_QUOTE)       ] = KEY_APOSTROPHE,
    [SIL_SDLK(SDLK_LEFTPAREN)   ] = KEY_LEFTPAREN,
    [SIL_SDLK(SDLK_RIGHTPAREN)  ] = KEY_RIGHTPAREN,
    [SIL_SDLK(SDLK_ASTERISK)    ] = KEY_ASTERISK,
    [SIL_SDLK(SDLK_PLUS)        ] = KEY_PLUS,
    [SIL_SDLK(SDLK_COMMA)       ] = KEY_COMMA,
    [SIL_SDLK(SDLK_MINUS)       ] = KEY_HYPHEN,
    [SIL_SDLK(SDLK_PERIOD)      ] = KEY_PERIOD,
    [SIL_SDLK(SDLK_SLASH)       ] = KEY_SLASH,
    [SIL_SDLK(SDLK_COLON)       ] = KEY_COLON,
    [SIL_SDLK(SDLK_SEMICOLON)   ] = KEY_SEMICOLON,
    [SIL_SDLK(SDLK_LESS)        ] = KEY_LESS,
    [SIL_SDLK(SDLK_EQUALS)      ] = KEY_EQUALS,
    [SIL_SDLK(SDLK_GREATER)     ] = KEY_GREATER,
    [SIL_SDLK(SDLK_QUESTION)    ] = KEY_QUESTION,
    [SIL_SDLK(SDLK_AT)          ] = KEY_ATSIGN,
    [SIL_SDLK(SDLK_LEFTBRACKET) ] = KEY_LEFTBRACKET,
    [SIL_SDLK(SDLK_BACKSLASH)   ] = KEY_BACKSLASH,
    [SIL_SDLK(SDLK_RIGHTBRACKET)] = KEY_RIGHTBRACKET,
    [SIL_SDLK(SDLK_CARET)       ] = KEY_CIRCUMFLEX,
    [SIL_SDLK(SDLK_UNDERSCORE)  ] = KEY_UNDERSCORE,
    [SIL_SDLK(SDLK_BACKQUOTE)   ] = KEY_BACKQUOTE,
#if !SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_EURO)        ] = KEY_EURO,
#endif
    [SIL_SDLK(SDLK_LSHIFT)      ] = KEY_LEFTSHIFT,
    [SIL_SDLK(SDLK_RSHIFT)      ] = KEY_RIGHTSHIFT,
    [SIL_SDLK(SDLK_LCTRL)       ] = KEY_LEFTCONTROL,
    [SIL_SDLK(SDLK_RCTRL)       ] = KEY_RIGHTCONTROL,
    [SIL_SDLK(SDLK_LALT)        ] = KEY_LEFTALT,
    [SIL_SDLK(SDLK_RALT)        ] = KEY_RIGHTALT,
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_LGUI)        ] = KEY_LEFTMETA,
    [SIL_SDLK(SDLK_RGUI)        ] = KEY_RIGHTMETA,
#else
    [SIL_SDLK(SDLK_LMETA)       ] = KEY_LEFTMETA,
    [SIL_SDLK(SDLK_RMETA)       ] = KEY_RIGHTMETA,
#endif
#if !SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_LSUPER)      ] = KEY_LEFTSUPER,
    [SIL_SDLK(SDLK_RSUPER)      ] = KEY_RIGHTSUPER,
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_NUMLOCKCLEAR)] = KEY_NUMLOCK,
#else
    [SIL_SDLK(SDLK_NUMLOCK)     ] = KEY_NUMLOCK,
#endif
    [SIL_SDLK(SDLK_CAPSLOCK)    ] = KEY_CAPSLOCK,
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_SCROLLLOCK)  ] = KEY_SCROLLLOCK,
#else
    [SIL_SDLK(SDLK_SCROLLOCK)   ] = KEY_SCROLLLOCK,
#endif
    [SIL_SDLK(SDLK_BACKSPACE)   ] = KEY_BACKSPACE,
    [SIL_SDLK(SDLK_UNDO)        ] = KEY_UNDO,
    [SIL_SDLK(SDLK_INSERT)      ] = KEY_INSERT,
    [SIL_SDLK(SDLK_DELETE)      ] = KEY_DELETE,
    [SIL_SDLK(SDLK_HOME)        ] = KEY_HOME,
    [SIL_SDLK(SDLK_END)         ] = KEY_END,
    [SIL_SDLK(SDLK_PAGEUP)      ] = KEY_PAGEUP,
    [SIL_SDLK(SDLK_PAGEDOWN)    ] = KEY_PAGEDOWN,
    [SIL_SDLK(SDLK_UP)          ] = KEY_UP,
    [SIL_SDLK(SDLK_DOWN)        ] = KEY_DOWN,
    [SIL_SDLK(SDLK_LEFT)        ] = KEY_LEFT,
    [SIL_SDLK(SDLK_RIGHT)       ] = KEY_RIGHT,
    [SIL_SDLK(SDLK_ESCAPE)      ] = KEY_ESCAPE,
    [SIL_SDLK(SDLK_F1)          ] = KEY_F1,
    [SIL_SDLK(SDLK_F2)          ] = KEY_F2,
    [SIL_SDLK(SDLK_F3)          ] = KEY_F3,
    [SIL_SDLK(SDLK_F4)          ] = KEY_F4,
    [SIL_SDLK(SDLK_F5)          ] = KEY_F5,
    [SIL_SDLK(SDLK_F6)          ] = KEY_F6,
    [SIL_SDLK(SDLK_F7)          ] = KEY_F7,
    [SIL_SDLK(SDLK_F8)          ] = KEY_F8,
    [SIL_SDLK(SDLK_F9)          ] = KEY_F9,
    [SIL_SDLK(SDLK_F10)         ] = KEY_F10,
    [SIL_SDLK(SDLK_F11)         ] = KEY_F11,
    [SIL_SDLK(SDLK_F12)         ] = KEY_F12,
    [SIL_SDLK(SDLK_F13)         ] = KEY_F13,
    [SIL_SDLK(SDLK_F14)         ] = KEY_F14,
    [SIL_SDLK(SDLK_F15)         ] = KEY_F15,
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_F16)         ] = KEY_F16,
    [SIL_SDLK(SDLK_F17)         ] = KEY_F17,
    [SIL_SDLK(SDLK_F18)         ] = KEY_F18,
    [SIL_SDLK(SDLK_F19)         ] = KEY_F19,
    [SIL_SDLK(SDLK_F20)         ] = KEY_F20,
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_PRINTSCREEN) ] = KEY_PRINTSCREEN,
#else
    [SIL_SDLK(SDLK_PRINT)       ] = KEY_PRINTSCREEN,
#endif
    [SIL_SDLK(SDLK_PAUSE)       ] = KEY_PAUSE,
    [SIL_SDLK(SDLK_MENU)        ] = KEY_MENU,
#if SDL_VERSION_ATLEAST(2,0,0)
    [SIL_SDLK(SDLK_KP_0)        ] = KEY_NUMPAD_0,
    [SIL_SDLK(SDLK_KP_1)        ] = KEY_NUMPAD_1,
    [SIL_SDLK(SDLK_KP_2)        ] = KEY_NUMPAD_2,
    [SIL_SDLK(SDLK_KP_3)        ] = KEY_NUMPAD_3,
    [SIL_SDLK(SDLK_KP_4)        ] = KEY_NUMPAD_4,
    [SIL_SDLK(SDLK_KP_5)        ] = KEY_NUMPAD_5,
    [SIL_SDLK(SDLK_KP_6)        ] = KEY_NUMPAD_6,
    [SIL_SDLK(SDLK_KP_7)        ] = KEY_NUMPAD_7,
    [SIL_SDLK(SDLK_KP_8)        ] = KEY_NUMPAD_8,
    [SIL_SDLK(SDLK_KP_9)        ] = KEY_NUMPAD_9,
#else
    [SIL_SDLK(SDLK_KP0)         ] = KEY_NUMPAD_0,
    [SIL_SDLK(SDLK_KP1)         ] = KEY_NUMPAD_1,
    [SIL_SDLK(SDLK_KP2)         ] = KEY_NUMPAD_2,
    [SIL_SDLK(SDLK_KP3)         ] = KEY_NUMPAD_3,
    [SIL_SDLK(SDLK_KP4)         ] = KEY_NUMPAD_4,
    [SIL_SDLK(SDLK_KP5)         ] = KEY_NUMPAD_5,
    [SIL_SDLK(SDLK_KP6)         ] = KEY_NUMPAD_6,
    [SIL_SDLK(SDLK_KP7)         ] = KEY_NUMPAD_7,
    [SIL_SDLK(SDLK_KP8)         ] = KEY_NUMPAD_8,
    [SIL_SDLK(SDLK_KP9)         ] = KEY_NUMPAD_9,
#endif
    [SIL_SDLK(SDLK_KP_PERIOD)   ] = KEY_NUMPAD_DOT,
    [SIL_SDLK(SDLK_KP_DIVIDE)   ] = KEY_NUMPAD_DIVIDE,
    [SIL_SDLK(SDLK_KP_MULTIPLY) ] = KEY_NUMPAD_MULTIPLY,
    [SIL_SDLK(SDLK_KP_MINUS)    ] = KEY_NUMPAD_SUBTRACT,
    [SIL_SDLK(SDLK_KP_PLUS)     ] = KEY_NUMPAD_ADD,
    [SIL_SDLK(SDLK_KP_EQUALS)   ] = KEY_NUMPAD_EQUALS,
    [SIL_SDLK(SDLK_KP_ENTER)    ] = KEY_NUMPAD_ENTER,
};  // sdlkey_map[]

/*-----------------------------------------------------------------------*/

/* Definitions of known joystick types. */

static const struct {
    /* Joystick name as reported by SDL. */
    const char *name;
    /* Button numbers for D-pad inputs, or -1 if none. */
    int8_t dpad_up, dpad_down, dpad_left, dpad_right;
    /* Mapping from logical to physical buttons. */
    int8_t button_map[INPUT_JOYBUTTON__NUM];
} joystick_defs[] = {

    {.name = "Microsoft X-Box 360 pad",
     .dpad_up = -1,
     .dpad_down = -1,
     .dpad_left = -1,
     .dpad_right = -1,
     .button_map = {
        [INPUT_JOYBUTTON_HOME      ] =  8,
        [INPUT_JOYBUTTON_START     ] =  6,
        [INPUT_JOYBUTTON_SELECT    ] =  7,
        [INPUT_JOYBUTTON_FACE_UP   ] =  3,
        [INPUT_JOYBUTTON_FACE_LEFT ] =  2,
        [INPUT_JOYBUTTON_FACE_RIGHT] =  1,
        [INPUT_JOYBUTTON_FACE_DOWN ] =  0,
        [INPUT_JOYBUTTON_L1        ] =  4,
        [INPUT_JOYBUTTON_R1        ] =  5,
        [INPUT_JOYBUTTON_L2        ] = -1,  // Analog only.
        [INPUT_JOYBUTTON_R2        ] = -1,  // Analog only.
        [INPUT_JOYBUTTON_L_STICK   ] =  9,
        [INPUT_JOYBUTTON_R_STICK   ] = 10,
    }},

    {.name = "Sony PLAYSTATION(R)3 Controller",
     .dpad_up = 4,
     .dpad_down = 6,
     .dpad_left = 7,
     .dpad_right = 5,
     .button_map = {
        [INPUT_JOYBUTTON_HOME      ] = 16,
        [INPUT_JOYBUTTON_START     ] =  3,
        [INPUT_JOYBUTTON_SELECT    ] =  0,
        [INPUT_JOYBUTTON_FACE_UP   ] = 12,
        [INPUT_JOYBUTTON_FACE_LEFT ] = 15,
        [INPUT_JOYBUTTON_FACE_RIGHT] = 13,
        [INPUT_JOYBUTTON_FACE_DOWN ] = 14,
        [INPUT_JOYBUTTON_L1        ] = 10,
        [INPUT_JOYBUTTON_R1        ] = 11,
        [INPUT_JOYBUTTON_L2        ] =  8,
        [INPUT_JOYBUTTON_R2        ] =  9,
        [INPUT_JOYBUTTON_L_STICK   ] =  1,
        [INPUT_JOYBUTTON_R_STICK   ] =  2,
    }},

};  // joystick_defs[]

/*-----------------------------------------------------------------------*/

/* Some function parameters are used only in SDL 1.2 or SDL 2.0+ builds. */

#if SDL_VERSION_ATLEAST(2,0,0)
# define UNUSED_SDL1  /*nothing*/
# define UNUSED_SDL2  UNUSED
#else
# define UNUSED_SDL1  UNUSED
# define UNUSED_SDL2  /*nothing*/
#endif

/*-----------------------------------------------------------------------*/

/* Event callback passed to sys_input_init(). */
static InputEventCallback event_callback;

/* Number of available joystick devices. */
static int num_joysticks;

/* Data for each joystick device. */
typedef struct SDLJoystickInfo SDLJoystickInfo;
struct SDLJoystickInfo {
    /* SDL joystick object. */
    SDL_Joystick *joystick;

    /* Button numbers for D-pad inputs, or -1 if none. */
    int dpad_up, dpad_down, dpad_left, dpad_right;
    /* Current state of D-pad button inputs. */
    uint8_t dpad_state_up, dpad_state_down, dpad_state_left, dpad_state_right;

    /* Mapping from logical to physical buttons. */
    int button_map[INPUT_JOYBUTTON__NUM];

#if SDL_VERSION_ATLEAST(2,0,0)
    /* Unique ID for this joystick. */
    SDL_JoystickID id;

    /* SDL force feedback object (NULL if not supported by the joystick). */
    SDL_Haptic *haptic;

    /* Current force feedback effect ID. */
    int rumble_effect;
#endif  // SDL >= 2.0.0
};
static SDLJoystickInfo *joysticks;

/* Joystick information array returned for sys_input_info(). */
static SysInputJoystick *joystick_info;

#ifdef SIL_PLATFORM_LINUX
/* Delayed key release state.  The value of each element is the SDL keycode
 * (SDLK_*) for the key if a delayed release was detected, 0 otherwise.
 * See KEYUP handling for why we need this. */
static int key_release[KEY__LAST];
#endif

#if SDL_VERSION_ATLEAST(2,0,0)
/* Last reported mouse position.  Needed by scroll events. */
static float mouse_last_x, mouse_last_y;
#endif

/* Mouse position override, for sys_input_mouse_set_position() (see comments
 * in that function). */
static uint8_t mouse_position_overridden;
static int mouse_position_override_x, mouse_position_override_y;

/* Text input flag. */
static uint8_t text_active;

/* Flag indicating whether a quit event (SDL_QUIT) has been received. */
static uint8_t quit_requested;

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/**
 * init_joystick:  Initialize the joystick data and (if requested) haptic
 * interface for the given joystick.
 *
 * [Parameters]
 *     joystick: SDLJoystickInfo structure of joystick to initialize.
 *     index: SDL joystick index.
 *     name: Joystick name reported by SDL.
 *     open_haptic: True to open a haptic interface (ignored with SDL 1.2).
 */
static void init_joystick(SDLJoystickInfo *joystick, int index,
                          const char *name, int open_haptic);

/**
 * normalize_joystick_axis:  Return the given SDL-style (signed 16-bit
 * integer) joystick axis value normalized to a floating-point value in
 * the range [-1,+1].
 *
 * [Parameters]
 *     raw: Raw axis value, an integer in [-32768,+32767].
 * [Return value]
 *     Normalized axis value, a floating-point value in [-1,+1].
 */
static float normalize_joystick_axis(int raw);

/**
 * convert_mouse_coords:  Convert mouse coordinates reported by SDL to
 * those expected by the SIL input subsystem.  This function takes into
 * account the position override set by sys_input_mouse_set_position().
 *
 * [Parameters]
 *     x_in, y_in: Integer mouse coordinates from SDL.
 *     x_out, y_out: Pointers to variables to receive the converted
 *         floating-point mouse coordinates.
 */
static void convert_mouse_coords(int x_in, int y_in,
                                 float *x_out, float *y_out);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_input_init(void (*event_callback_)(const struct InputEvent *))
{
    PRECOND(event_callback_ != NULL, return 0);

    event_callback = event_callback_;

#if !SDL_VERSION_ATLEAST(2,0,0)
    SDL_EnableUNICODE(0);
    SDL_EnableKeyRepeat(0, 0);
#endif

    if (UNLIKELY(SDL_InitSubSystem(SDL_INIT_JOYSTICK) != 0)) {
        DLOG("SDL_InitSubSystem(SDL_INIT_JOYSTICK) failed: %s",
             SDL_GetError());
        return 0;
    }

    SDL_JoystickEventState(SDL_ENABLE);
    num_joysticks = SDL_NumJoysticks();
    if (num_joysticks > 0) {
        joysticks = mem_alloc(sizeof(*joysticks) * num_joysticks, 0, 0);
        joystick_info = mem_alloc(sizeof(*joystick_info) * num_joysticks, 0, 0);
        if (UNLIKELY(!joysticks) || UNLIKELY(!joystick_info)) {
            DLOG("No memory for joystick info");
            mem_free(joysticks);
            joysticks = NULL;
            mem_free(joystick_info);
            joystick_info = NULL;
            SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
            return 0;
        }
    }
    int opened_haptic = 0;
    for (int i = 0; i < num_joysticks; i++) {
        joysticks[i].joystick = SDL_JoystickOpen(i);
        if (UNLIKELY(!joysticks[i].joystick)) {
            DLOG("Failed to open joystick %d: %s", i, SDL_GetError());
        } else {
#if SDL_VERSION_ATLEAST(2,0,0)
            const char *name = SDL_JoystickName(joysticks[i].joystick);
#else
            const char *name = SDL_JoystickName(i);
#endif
            ASSERT(name != NULL, name = "(null)");
            DLOG("Joystick %d: %s", i, name);
#if SDL_VERSION_ATLEAST(2,0,0)
            if (SDL_JoystickIsHaptic(joysticks[i].joystick) > 0
             && !opened_haptic) {
                if (SDL_InitSubSystem(SDL_INIT_HAPTIC) != 0) {
                    DLOG("SDL_InitSubSystem(SDL_INIT_HAPTIC) failed: %s",
                         SDL_GetError());
                } else {
                    opened_haptic = 1;
                }
            }
#endif
            init_joystick(&joysticks[i], i, name, opened_haptic);
        }
    }

#if SDL_VERSION_ATLEAST(2,0,0)
    mouse_last_x = mouse_last_y = 0.5;
#endif
    mouse_position_overridden = 0;
    text_active = 0;
    quit_requested = 0;

    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_input_cleanup(void)
{
    for (int i = 0; i < num_joysticks; i++) {
        if (joysticks[i].joystick) {
#if SDL_VERSION_ATLEAST(2,0,0)
            if (joysticks[i].haptic) {
                SDL_HapticClose(joysticks[i].haptic);
            }
#endif
            SDL_JoystickClose(joysticks[i].joystick);
        }
    }
    mem_free(joysticks);
    joysticks = NULL;
    mem_free(joystick_info);
    joystick_info = NULL;
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_QuitSubSystem(SDL_INIT_HAPTIC);
#endif
    SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
}

/*-----------------------------------------------------------------------*/

void sys_input_update(void)
{
    const double now = time_now();

#ifdef SIL_PLATFORM_LINUX
    /* Array of keys newly pressed during this call.  See KEYUP handling
     * for why we need this. */
    uint8_t newkeys[KEY__LAST];
    mem_clear(newkeys, sizeof(newkeys));
    for (int key = 1; key < lenof(key_release); key++) {
        if (key_release[key]) {
            (*event_callback)(&(InputEvent){
                .type = INPUT_EVENT_KEYBOARD, .detail = INPUT_KEYBOARD_KEY_UP,
                .timestamp = now,
                {.keyboard = {.key = key, .system_key = key_release[key]}}});
            key_release[key] = 0;
        }
    }
#endif

    SDL_PumpEvents();

    /* Some versions of SDL crash on SDL_JoystickUpdate() when there are no
     * joysticks connected. */
    if (num_joysticks > 0) {
        SDL_JoystickUpdate();
    }

    SDL_Event sdl_event;
    while (SDL_PollEvent(&sdl_event)) {
        switch (sdl_event.type) {

          case SDL_JOYAXISMOTION: {
            if (UNLIKELY(sdl_event.jaxis.which >= num_joysticks)) {
                DLOG("Invalid joystick device index %d",
                     sdl_event.jaxis.which);
                break;
            }
            SDLJoystickInfo *info = &joysticks[sdl_event.jaxis.which];
            float x = normalize_joystick_axis(sdl_event.jaxis.value);
            float y;
            if (sdl_event.jaxis.axis + 1 >= SDL_JoystickNumAxes(
                    info->joystick)) {
                y = 0;
            } else {
                y = normalize_joystick_axis(
                    SDL_JoystickGetAxis(
                        info->joystick,
                        sdl_event.jaxis.axis + 1));
            }
            (*event_callback)(&(InputEvent){
                .type = INPUT_EVENT_JOYSTICK,
                .detail = INPUT_JOYSTICK_STICK_CHANGE,
                .timestamp = now,
                {.joystick = {.device = sdl_event.jaxis.which,
                              .index = sdl_event.jaxis.axis,
                              .x = x, .y = y}}});
            if (sdl_event.jaxis.axis > 0) {
                y = x;
                x = normalize_joystick_axis(
                    SDL_JoystickGetAxis(
                        info->joystick,
                        sdl_event.jaxis.axis - 1));
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_JOYSTICK,
                    .detail = INPUT_JOYSTICK_STICK_CHANGE,
                    .timestamp = now,
                    {.joystick = {.device = sdl_event.jaxis.which,
                                  .index = sdl_event.jaxis.axis - 1,
                                  .x = x, .y = y}}});
            }
            break;
          }  // case SDL_JOYAXISMOTION

          case SDL_JOYBUTTONDOWN:
          case SDL_JOYBUTTONUP: {
            if (UNLIKELY(sdl_event.jbutton.which >= num_joysticks)) {
                DLOG("Invalid joystick device index %d",
                     sdl_event.jbutton.which);
                break;
            }
            SDLJoystickInfo *info = &joysticks[sdl_event.jbutton.which];
            (*event_callback)(&(InputEvent){
                .type = INPUT_EVENT_JOYSTICK,
                .detail = (sdl_event.type == SDL_JOYBUTTONDOWN
                           ? INPUT_JOYSTICK_BUTTON_DOWN
                           : INPUT_JOYSTICK_BUTTON_UP),
                .timestamp = now,
                {.joystick = {.device = sdl_event.jbutton.which,
                              .index = sdl_event.jbutton.button}}});
            int is_dpad = 0;
            if (sdl_event.jbutton.button == info->dpad_up) {
                is_dpad = 1;
                info->dpad_state_up = (sdl_event.type == SDL_JOYBUTTONDOWN);
            } else if (sdl_event.jbutton.button == info->dpad_down) {
                is_dpad = 1;
                info->dpad_state_down = (sdl_event.type == SDL_JOYBUTTONDOWN);
            } else if (sdl_event.jbutton.button == info->dpad_left) {
                is_dpad = 1;
                info->dpad_state_left = (sdl_event.type == SDL_JOYBUTTONDOWN);
            } else if (sdl_event.jbutton.button == info->dpad_right) {
                is_dpad = 1;
                info->dpad_state_right = (sdl_event.type == SDL_JOYBUTTONDOWN);
            }
            if (is_dpad) {
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_JOYSTICK,
                    .detail = INPUT_JOYSTICK_DPAD_CHANGE,
                    .timestamp = now,
                    {.joystick = {
                        .device = sdl_event.jbutton.which,
                        .x = info->dpad_state_right - info->dpad_state_left,
                        .y = info->dpad_state_down - info->dpad_state_up}}});
            }
            break;
          }  // case SDL_JOYBUTTONDOWN, SDL_JOYBUTTONUP

          case SDL_JOYHATMOTION: {
            if (UNLIKELY(sdl_event.jhat.which >= num_joysticks)) {
                DLOG("Invalid joystick device index %d",
                     sdl_event.jhat.which);
                break;
            }
            if (sdl_event.jhat.hat != 0) {
                break;  // If more than one hat, assume hat 0 is the D-pad.
            }
            const int x = ((sdl_event.jhat.value & SDL_HAT_LEFT) ? -1 :
                           (sdl_event.jhat.value & SDL_HAT_RIGHT) ? +1 : 0);
            const int y = ((sdl_event.jhat.value & SDL_HAT_UP) ? -1 :
                           (sdl_event.jhat.value & SDL_HAT_DOWN) ? +1 : 0);
            (*event_callback)(&(InputEvent){
                .type = INPUT_EVENT_JOYSTICK,
                .detail = INPUT_JOYSTICK_DPAD_CHANGE,
                .timestamp = now,
                {.joystick = {.device = sdl_event.jbutton.which,
                              .x = x, .y = y}}});
            break;
          }  // case SDL_JOYHATMOTION

          case SDL_KEYDOWN: {
            const unsigned int keysym = sdl_event.key.keysym.sym;
            const int key = (SIL_SDLK(keysym) < lenof(sdlkey_map)
                             ? sdlkey_map[SIL_SDLK(keysym)] : 0);
#ifdef SIL_PLATFORM_MACOSX
            /* Catch Command-Q and treat as a quit event, since SDL traps
             * these when keyboard input is grabbed. */
            if (key == KEY_Q && (sdl_event.key.keysym.mod & KMOD_GUI)) {
                quit_requested = 1;
                break;
            }
#endif
            if (key) {
#ifdef SIL_PLATFORM_LINUX
                key_release[key] = 0;
                newkeys[key] = 1;
#endif
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_KEYBOARD,
                    .detail = INPUT_KEYBOARD_KEY_DOWN,
                    .timestamp = now,
                    {.keyboard = {.key = key,
                                  .system_key = sdl_event.key.keysym.sym}}});
            } else {
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_KEYBOARD,
                    .detail = INPUT_KEYBOARD_SYSTEM_KEY_DOWN,
                    .timestamp = now,
                    {.keyboard = {.key = KEY__NONE,
                                  .system_key = sdl_event.key.keysym.sym}}});
            }
            if (text_active) {
                InputEvent text_event = {
                    .type = INPUT_EVENT_TEXT, .timestamp = now};
                int ignore = 0;
                if (sdl_event.key.keysym.sym == SDLK_BACKSPACE) {
                    text_event.detail = INPUT_TEXT_BACKSPACE;
                } else if (sdl_event.key.keysym.sym == SDLK_DELETE) {
                    text_event.detail = INPUT_TEXT_DELETE;
                } else if (sdl_event.key.keysym.sym == SDLK_LEFT) {
                    text_event.detail = INPUT_TEXT_CURSOR_LEFT;
                } else if (sdl_event.key.keysym.sym == SDLK_RIGHT) {
                    text_event.detail = INPUT_TEXT_CURSOR_RIGHT;
                } else if (sdl_event.key.keysym.sym == SDLK_HOME) {
                    text_event.detail = INPUT_TEXT_CURSOR_HOME;
                } else if (sdl_event.key.keysym.sym == SDLK_END) {
                    text_event.detail = INPUT_TEXT_CURSOR_END;
                } else if (sdl_event.key.keysym.sym == SDLK_ESCAPE) {
                    text_event.detail = INPUT_TEXT_CANCELLED;
                } else if (sdl_event.key.keysym.sym == SDLK_RETURN
                        || sdl_event.key.keysym.sym == SDLK_KP_ENTER) {
                    text_event.detail = INPUT_TEXT_DONE;
#if !SDL_VERSION_ATLEAST(2,0,0)
                } else if (sdl_event.key.keysym.unicode >= ' ') {
                    text_event.detail = INPUT_TEXT_INPUT;
                    text_event.text.ch = (int32_t)sdl_event.key.keysym.unicode;
#endif
                } else {
                    ignore = 1;
                }
                if (!ignore) {
                    (*event_callback)(&text_event);
                }
            }
            break;
          }  // case SDL_KEYDOWN

          case SDL_KEYUP: {
            const unsigned int keysym = sdl_event.key.keysym.sym;
            const int key = (SIL_SDLK(keysym) < lenof(sdlkey_map)
                             ? sdlkey_map[SIL_SDLK(keysym)] : 0);
            if (key) {
#ifdef SIL_PLATFORM_LINUX
                /*
                 * Some broken(?) X servers seem to send KeyRelease events
                 * immediately after the corresponding KeyPress for keys
                 * with modifiers pressed, like Ctrl+A, rather than sending
                 * KeyRelease followed by KeyPress at each repeat interval.
                 * We can't reliably recover the physical key state from
                 * these bogus events, but we can at least make sure that
                 * the keypress isn't lost completely by delaying the
                 * release event until the next sys_input_update() call.
                 */
                if (newkeys[key]) {
                    key_release[key] = sdl_event.key.keysym.sym;
                    break;
                }
                key_release[key] = 0;
                newkeys[key] = 0;
#endif
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_KEYBOARD,
                    .detail = INPUT_KEYBOARD_KEY_UP,
                    .timestamp = now,
                    {.keyboard = {.key = key,
                                  .system_key = sdl_event.key.keysym.sym}}});
            } else {
                (*event_callback)(&(InputEvent){
                    .type = INPUT_EVENT_KEYBOARD,
                    .detail = INPUT_KEYBOARD_SYSTEM_KEY_UP,
                    .timestamp = now,
                    {.keyboard = {.key = KEY__NONE,
                                  .system_key = sdl_event.key.keysym.sym}}});
            }
            break;
          }  // case SDL_KEYUP

          case SDL_MOUSEBUTTONDOWN:
          case SDL_MOUSEBUTTONUP: {
            const int is_down = (sdl_event.type == SDL_MOUSEBUTTONDOWN);
            InputEventDetail detail = 0;
            int scroll = 0;
            switch (sdl_event.button.button) {
              case SDL_BUTTON_LEFT:
                detail = is_down ? INPUT_MOUSE_LMB_DOWN : INPUT_MOUSE_LMB_UP;
                break;
              case SDL_BUTTON_MIDDLE:
                detail = is_down ? INPUT_MOUSE_MMB_DOWN : INPUT_MOUSE_MMB_UP;
                break;
              case SDL_BUTTON_RIGHT:
                detail = is_down ? INPUT_MOUSE_RMB_DOWN : INPUT_MOUSE_RMB_UP;
                break;
#if !SDL_VERSION_ATLEAST(2,0,0)
              case SDL_BUTTON_WHEELUP:
                if (is_down) {
                    detail = INPUT_MOUSE_SCROLL_V;
                    scroll = -1;
                }
                break;
              case SDL_BUTTON_WHEELDOWN:
                if (is_down) {
                    detail = INPUT_MOUSE_SCROLL_V;
                    scroll = 1;
                }
                break;
#endif
            }
            if (detail) {
                InputEvent event = {
                    .type = INPUT_EVENT_MOUSE, .detail = detail,
                    .timestamp = now};
                convert_mouse_coords(sdl_event.button.x, sdl_event.button.y,
                                     &event.mouse.x, &event.mouse.y);
                event.mouse.scroll = scroll;
                (*event_callback)(&event);
#if SDL_VERSION_ATLEAST(2,0,0)
                mouse_last_x = event.mouse.x;
                mouse_last_y = event.mouse.y;
#endif
            }
            break;
          }  // case SDL_MOUSEBUTTONDOWN, SDL_MOUSEBUTTONUP

          case SDL_MOUSEMOTION: {
            if (mouse_position_overridden) {
                mouse_position_overridden =
                    (sdl_event.motion.x != mouse_position_override_x ||
                     sdl_event.motion.y != mouse_position_override_y);
            } else {
                InputEvent event = {
                    .type = INPUT_EVENT_MOUSE, .detail = INPUT_MOUSE_MOVE,
                    .timestamp = now};
                convert_mouse_coords(sdl_event.motion.x, sdl_event.motion.y,
                                     &event.mouse.x, &event.mouse.y);
                (*event_callback)(&event);
#if SDL_VERSION_ATLEAST(2,0,0)
                mouse_last_x = event.mouse.x;
                mouse_last_y = event.mouse.y;
#endif
            }
            break;
          }  // case SDL_MOUSEMOTION

#if SDL_VERSION_ATLEAST(2,0,0)
          case SDL_MOUSEWHEEL: {
            InputEvent event = {
                .type = INPUT_EVENT_MOUSE, .timestamp = now,
                {.mouse = {.x = mouse_last_x, .y = mouse_last_y}}};
            if (sdl_event.wheel.x != 0) {
                event.detail = INPUT_MOUSE_SCROLL_H;
                event.mouse.scroll = sdl_event.wheel.x;
                (*event_callback)(&event);
            }
            if (sdl_event.wheel.y != 0) {
                event.detail = INPUT_MOUSE_SCROLL_V;
                event.mouse.scroll = sdl_event.wheel.y;
                (*event_callback)(&event);
            }
            break;
          }  // case SDL_MOUSEWHEEL
#endif

          case SDL_QUIT:
            quit_requested = 1;
            break;

#if SDL_VERSION_ATLEAST(2,0,0)
          case SDL_TEXTINPUT: {
            InputEvent text_event = {
                .type = INPUT_EVENT_TEXT, .detail = INPUT_TEXT_INPUT,
                .timestamp = now};
            const char *text = sdl_event.text.text;
            while ((text_event.text.ch = utf8_read(&text)) != 0) {
                if (UNLIKELY(text_event.text.ch < 0)) {
                    /* This should never occur unless SDL is broken. */
                    DLOG("Invalid UTF-8 in SDL input string");
                    continue;
                }
                (*event_callback)(&text_event);
            }
          }  // case SDL_TEXTINPUT
#endif

        }  // switch (sdl_event.type)
    }  // while (SDL_PollEvent(&event))
}

/*-----------------------------------------------------------------------*/

void sys_input_info(SysInputInfo *info_ret)
{
    info_ret->has_joystick = (num_joysticks > 0);
    info_ret->num_joysticks = num_joysticks;
    info_ret->joysticks = joystick_info;
    for (int i = 0; i < num_joysticks; i++) {
        joystick_info[i].connected   = 1;
#if SDL_VERSION_ATLEAST(2,0,0)
        joystick_info[i].can_rumble  = (joysticks[i].haptic != NULL);
#else
        joystick_info[i].can_rumble  = 0;
#endif
        joystick_info[i].num_buttons =
            SDL_JoystickNumButtons(joysticks[i].joystick);
        joystick_info[i].num_sticks  =
            SDL_JoystickNumAxes(joysticks[i].joystick);
    }

    /* SDL doesn't seem to provide a way to check for the presence of
     * keyboards or mice, so assume they're present. */
    info_ret->has_keyboard = 1;
    info_ret->keyboard_is_full = 1;
    info_ret->has_mouse = 1;

    /* We use SDL's key translation to provide text entry. */
    info_ret->has_text = 1;
    info_ret->text_uses_custom_interface = 0;
    info_ret->text_has_prompt = 0;

    /* We don't support touch events under SDL. */
    info_ret->has_touch = 0;
}

/*-----------------------------------------------------------------------*/

void sys_input_grab(int grab)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_SetWindowGrab(sdl_window(), grab);
#else
    SDL_WM_GrabInput(grab ? SDL_GRAB_ON : SDL_GRAB_OFF);
#endif
}

/*-----------------------------------------------------------------------*/

int sys_input_is_quit_requested(void)
{
    return quit_requested;
}

/*-----------------------------------------------------------------------*/

int sys_input_is_suspend_requested(void)
{
    /* Not supported under SDL. */
    return 0;
}

/*-----------------------------------------------------------------------*/

void sys_input_acknowledge_suspend_request(void)
{
    /* Not supported under SDL. */
}

/*************************************************************************/
/********************* Interface: Joystick handling **********************/
/*************************************************************************/

int sys_input_joystick_button_mapping(int index, int name)
{
    return joysticks[index].button_map[name];
}

/*-----------------------------------------------------------------------*/

void sys_input_joystick_rumble(UNUSED_SDL1 int index, UNUSED_SDL1 float left,
                               UNUSED_SDL1 float right, UNUSED_SDL1 float time)
{
#if SDL_VERSION_ATLEAST(2,0,0)

    if (joysticks[index].rumble_effect != -1) {
        SDL_HapticDestroyEffect(joysticks[index].haptic,
                                joysticks[index].rumble_effect);
        joysticks[index].rumble_effect = -1;
    }

    SDL_HapticEffect effect;
    mem_clear(&effect, sizeof(effect));
    /* SDL doesn't (as of HG r4865) have a plain "rumble" effect, so we
     * fake it with a sine wave of zero amplitude by setting magnitude
     * and offset equal. */
    effect.type = SDL_HAPTIC_SINE;
    effect.periodic.magnitude = iroundf((left+right)/2 * 32767);
    effect.periodic.offset = effect.periodic.magnitude;
    effect.periodic.length = iroundf(time * 1000);
    joysticks[index].rumble_effect =
        SDL_HapticNewEffect(joysticks[index].haptic, &effect);
    if (joysticks[index].rumble_effect == -1) {
        DLOG("Failed to create rumble effect: %s", SDL_GetError());
        return;
    }

    if (SDL_HapticRunEffect(joysticks[index].haptic,
                            joysticks[index].rumble_effect, 1) != 0) {
        DLOG("Failed to start rumble effect: %s", SDL_GetError());
        SDL_HapticDestroyEffect(joysticks[index].haptic,
                                joysticks[index].rumble_effect);
        joysticks[index].rumble_effect = -1;
    }

#endif  // SDL >= 2.0.0
}

/*************************************************************************/
/*********************** Interface: Mouse handling ***********************/
/*************************************************************************/

void sys_input_mouse_set_position(float x, float y)
{
    const int width = graphics_display_width();
    const int height = graphics_display_height();
    if (width != 0 && height != 0) {
        const int ix = bound(iroundf(x*width),  0, width-1);
        const int iy = bound(iroundf(y*height), 0, height-1);
        int curx, cury;
        SDL_GetMouseState(&curx, &cury);
        if (ix != curx || iy != cury) {
#if SDL_VERSION_ATLEAST(2,0,0)
            SDL_WarpMouseInWindow(sdl_window(), ix, iy);
#else
            SDL_WarpMouse(ix, iy);
#endif
            (*event_callback)(&(InputEvent){
                .type = INPUT_EVENT_MOUSE, .detail = INPUT_MOUSE_MOVE,
                .timestamp = time_now(),
                .mouse = {.x = (float)ix / (float)width,
                          .y = (float)iy / (float)height}});
            /* There may already be mouse events in the queue, so until the
             * SDL_MOUSEMOTION event generated by this call is processed,
             * we override the mouse position to the requested one. */
            mouse_position_overridden = 1;
            mouse_position_override_x = ix;
            mouse_position_override_y = iy;
        }
    }
}

/*************************************************************************/
/******************** Interface: Text entry handling *********************/
/*************************************************************************/

void sys_input_text_set_state(int on, UNUSED const char *text,
                              UNUSED const char *prompt)
{
    text_active = (on != 0);
#if !SDL_VERSION_ATLEAST(2,0,0)
    if (on) {
        SDL_EnableUNICODE(1);
        SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,
                            SDL_DEFAULT_REPEAT_INTERVAL);
    } else {
        SDL_EnableUNICODE(0);
        SDL_EnableKeyRepeat(0, 0);
    }
#endif
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void init_joystick(SDLJoystickInfo *joystick, UNUSED_SDL1 int index,
                          const char *name, UNUSED_SDL1 int open_haptic)
{
    ASSERT(name != NULL);

    for (int i = 0; i < lenof(joystick->button_map); i++) {
        joystick->button_map[i] = -1;
    }
    joystick->dpad_up = -1;
    joystick->dpad_down = -1;
    joystick->dpad_left = -1;
    joystick->dpad_right = -1;
    joystick->dpad_state_up = 0;
    joystick->dpad_state_down = 0;
    joystick->dpad_state_left = 0;
    joystick->dpad_state_right = 0;

    UNUSED_SDL1 int found_map = 0;
    for (int i = 0; i < lenof(joystick_defs); i++) {
        if (strcmp(name, joystick_defs[i].name) == 0) {
            for (int j = 0; j < INPUT_JOYBUTTON__NUM; j++) {
                joystick->button_map[j] = joystick_defs[i].button_map[j];
            }
            joystick->dpad_up = joystick_defs[i].dpad_up;
            joystick->dpad_down = joystick_defs[i].dpad_down;
            joystick->dpad_left = joystick_defs[i].dpad_left;
            joystick->dpad_right = joystick_defs[i].dpad_right;
            found_map = 1;
            break;
        }
    }

#if SDL_VERSION_ATLEAST(2,0,0)
    joystick->id = SDL_JoystickInstanceID(joystick->joystick);

    if (SDL_IsGameController(index) && !found_map) {
        SDL_GameController *controller = SDL_GameControllerOpen(index);
        if (UNLIKELY(controller)) {
            DLOG("Failed to open game controller interface for joystick %d:"
                 " %s", index, SDL_GetError());
        } else {
            static const int buttons[] = {
                [INPUT_JOYBUTTON_HOME]       = SDL_CONTROLLER_BUTTON_GUIDE,
                [INPUT_JOYBUTTON_START]      = SDL_CONTROLLER_BUTTON_START,
                [INPUT_JOYBUTTON_SELECT]     = SDL_CONTROLLER_BUTTON_BACK,
                [INPUT_JOYBUTTON_FACE_UP]    = SDL_CONTROLLER_BUTTON_Y,
                [INPUT_JOYBUTTON_FACE_LEFT]  = SDL_CONTROLLER_BUTTON_X,
                [INPUT_JOYBUTTON_FACE_RIGHT] = SDL_CONTROLLER_BUTTON_B,
                [INPUT_JOYBUTTON_FACE_DOWN]  = SDL_CONTROLLER_BUTTON_A,
                [INPUT_JOYBUTTON_L1]         = SDL_CONTROLLER_BUTTON_LEFTSHOULDER,
                [INPUT_JOYBUTTON_R1]         = SDL_CONTROLLER_BUTTON_RIGHTSHOULDER,
                [INPUT_JOYBUTTON_L2]         = -1,  // Not exported by SDL!
                [INPUT_JOYBUTTON_R2]         = -1,  // Not exported by SDL!
                [INPUT_JOYBUTTON_L_STICK]    = SDL_CONTROLLER_BUTTON_LEFTSTICK,
                [INPUT_JOYBUTTON_R_STICK]    = SDL_CONTROLLER_BUTTON_RIGHTSTICK,
            };
            for (int i = 0; i < lenof(buttons); i++) {
                if (buttons[i] < 0) {
                    continue;
                }
                SDL_GameControllerButtonBind bind =
                    SDL_GameControllerGetBindForButton(controller, buttons[i]);
                if (bind.bindType == SDL_CONTROLLER_BINDTYPE_BUTTON) {
                    joystick->button_map[i] = bind.value.button;
                }
            }

            SDL_GameControllerButtonBind bind;
            bind = SDL_GameControllerGetBindForButton(
                controller, SDL_CONTROLLER_BUTTON_DPAD_UP);
            if (bind.bindType == SDL_CONTROLLER_BINDTYPE_BUTTON) {
                joystick->dpad_up = bind.value.button;
            }
            bind = SDL_GameControllerGetBindForButton(
                controller, SDL_CONTROLLER_BUTTON_DPAD_DOWN);
            if (bind.bindType == SDL_CONTROLLER_BINDTYPE_BUTTON) {
                joystick->dpad_down = bind.value.button;
            }
            bind = SDL_GameControllerGetBindForButton(
                controller, SDL_CONTROLLER_BUTTON_DPAD_LEFT);
            if (bind.bindType == SDL_CONTROLLER_BINDTYPE_BUTTON) {
                joystick->dpad_left = bind.value.button;
            }
            bind = SDL_GameControllerGetBindForButton(
                controller, SDL_CONTROLLER_BUTTON_DPAD_RIGHT);
            if (bind.bindType == SDL_CONTROLLER_BINDTYPE_BUTTON) {
                joystick->dpad_right = bind.value.button;
            }

            SDL_GameControllerClose(controller);
        }
    }

    if (open_haptic) {
        joystick->haptic = SDL_HapticOpenFromJoystick(joystick->joystick);
        if (!joystick->haptic) {
            DLOG("Failed to get force feedback handle for joystick %d: %s",
                 index, SDL_GetError());
        } else {
            joystick->rumble_effect = -1;
        }
    }
#endif
}

/*-----------------------------------------------------------------------*/

static float normalize_joystick_axis(int raw)
{
    if (raw < 0) {
        return raw / 32768.0f;
    } else {
        return raw / 32767.0f;
    }
}

/*-----------------------------------------------------------------------*/

static void convert_mouse_coords(int x_in, int y_in,
                                 float *x_out, float *y_out)
{
    if (mouse_position_overridden) {
        x_in = mouse_position_override_x;
        y_in = mouse_position_override_y;
    }

    const int width = graphics_display_width();
    const int height = graphics_display_height();
    if (LIKELY(width != 0 && height != 0)) {
        *x_out = (float)bound(x_in, 0, width-1)  / (float)width;
        *y_out = (float)bound(y_in, 0, height-1) / (float)height;
    } else {
        /* Since SDL doesn't process events without a window open, we
         * should never hit this branch, but default to the center of the
         * screen if we don't have a window. */
        *x_out = *y_out = 0.5;  // NOTREACHED
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
